import cv2 as cv
import numpy as np
import utils

# Equalize histogram to fill spectrum of luminosity
def flattenHistogram(img):
    # Convert BGR image to LAB image (Luminosity, A-channel, B-channel)
    img_lab = cv.cvtColor(img, cv.COLOR_BGR2LAB)

    # Split channels
    l, a, b = cv.split(img_lab)

    # Old method
    # equalized = cv.equalizeHist(l)
    # merged_img = cv.merge((equalized, a, b))

    # CLAHE algorithm
    clahe = cv.createCLAHE(clipLimit=3.0, tileGridSize=(8, 8))
    clahe_img = clahe.apply(l)

    merged_img = cv.merge((clahe_img, a, b))

    return cv.cvtColor(merged_img, cv.COLOR_LAB2BGR)


# Detect edges using canny edge detection
# Parameters for edge detection is automatically optimized based
# on lighting conditions, contrast etc.
def cannyEdgeDetecion(img, sigma=0.33):
    # Find the median of the all pixel intensities in the image
    # median = np.median(img)

    # Automaticallly choose lower and upper boundries based on median
    # lower = int(max(0, (1.0 - sigma) * median))
    # upper = int(max(255, (1.0 + sigma) * median))

    # Convert to grayscale
    gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)

    median = cv.medianBlur(gray, 3)

    # Apply threshold
    thresh = cv.threshold(median, 150, 255, cv.THRESH_BINARY)[1]

    # Dilated mask - mask = morphology edgeout
    kernel = cv.getStructuringElement(cv.MORPH_RECT, (5, 5))
    dilate = cv.morphologyEx(thresh, cv.MORPH_DILATE, kernel)
    # dilate = cv.dilate(thresh, kernel, iterations=1)

    # Get absolute difference between dilaton and threshold
    diff = cv.absdiff(dilate, thresh)

    edges = 255 - diff

    # Blur the image using median blurring to remove noise
    # blurred_img =  cv.GaussianBlur(img, (3,3), 0)

    # Detect edges using Canny in OpenCV
    # edge = cv.Canny(thresh)

    return edges


# DOES THIS WORK????
def detectCorners(img, img2):
    copyImg = img.copy()
    # img_new = cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    # img_new = np.float32(img)
    corners = cv.goodFeaturesToTrack(copyImg, 200, 0.01, 10)

    if corners is not None:
        corners = np.int0(corners)

        for corner in corners:
            x, y = corner.ravel()
            cv.circle(img2, (x, y), 5, 255, -1)

    return img2


def findKeyPoints(frame, method="ORB", numOfKeypoints=100):
    # Check if method is a valid value
    methods = ["ORB", "SIFT", "FAST"]
    if method not in methods:
        raise ValueError("Invalid method type. Expected one of: %s" % methods)

    # Convert frame to grayscale
    frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)

    # ORB - Oriented FASTand Rotated BRIEF
    if method == methods[0]:
        orb = cv.ORB_create(numOfKeypoints)
        kp, des = orb.detectAndCompute(frame_gray, None)

        # Convert to CV_32F format
        des = np.float32(des)
        return kp, des

    # SIFT - Scale-Invariant Feature Transform
    elif method == methods[1]:
        sift = cv.SIFT_create(numOfKeypoints)
        return sift.detectAndCompute(frame_gray, None)

    elif method == methods[2]:
        # Find keypoints with FAST
        fast = cv.FastFeatureDetector_create()
        kp = fast.detect(frame_gray, None)

        # Find descriptors with BRIEF
        brief = cv.xfeatures2d.BriefDescriptorExtractor_create()
        kp, des = brief.compute(frame_gray, kp)
        return kp, des


def findMatches(img1, img2, kp1, kp2, des1, des2):
    # FLANN parameters
    FLANN_INDEX_KDTREE = 0
    FLANN_INDEX_LSH = 1
    RATIO_THRESHOLD = 0.7

    # index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=1)
    index_params = dict(
        algorithm=FLANN_INDEX_LSH,
        table_number=10,  # 12
        key_size=20,  # 20
        multi_probe_level=2,
    )  # 2
    search_params = dict(checks=10)  # or pass empty dictionary

    flann = cv.FlannBasedMatcher(index_params, search_params)

    matches = flann.knnMatch(des1, des2, k=2)

    # bf = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)
    # matches = bf.match(des1, des2)

    goodMatches = []

    # Filter matches using Lowe's ratio test
    for m in matches:
        if m[0].distance < RATIO_THRESHOLD * m[1].distance:
            goodMatches.append(m)

    return goodMatches

    # for m in matches:
    #     print(m._distance)

    # print(kp1[0].pt)

    # # Distance between descriptors. The lower, the better it is.
    # print(matches[0][0].trainIdx, matches[0][0].queryIdx, matches[0][0].imgIdx)
    # print(matches[0][1].trainIdx, matches[0][1].queryIdx, matches[0][1].imgIdx)

    # distanceBetweenMatches(matches, kp1, kp2)

    # draw_params = dict(
    #     matchColor=(0, 255, 0),
    #     singlePointColor=(0, 0, 255),
    #     matchesMask=matchesMask,
    #     flags=0,
    # )

    # img3 = cv.drawMatchesKnn(img1, kp1, img2, kp2, matches, None, **draw_params)

    # cv.imshow("frame", utils.rescaleFrame(img3, 0.4))
    # cv.waitKey(0)


def distanceBetweenMatches(matches, kp1, kp2):
    if len(matches) == 0:
        return 0, 0

    avgDistanceY = 0
    avgDistanceX = 0

    # print(kp1[19].pt)
    # print(kp2[19].pt)
    # print(dir(matches[19][0]))
    # print((matches[0][0].distance))

    for m in matches:
        # Get indexes for keypoints from match object
        index1 = m[0].queryIdx
        index2 = m[0].trainIdx

        # Calculate distances between keypoints based on index
        avgDistanceX += abs((kp2[index2].pt[0] - kp1[index1].pt[0]))
        avgDistanceY += abs(kp2[index2].pt[1] - kp1[index1].pt[1])

    avgDistanceY = avgDistanceY / len(matches)
    avgDistanceX = avgDistanceX / len(matches)

    print(avgDistanceX, avgDistanceY)
    return avgDistanceY, avgDistanceX


# Testing
cap = cv.VideoCapture("./Resources/fouling.mp4")
cap.set(cv.CAP_PROP_POS_FRAMES, 1100)

_, img1 = cap.read()

# for i in range(30):
#     cap.read
_, img2 = cap.read()

kp1, des1 = findKeyPoints(img1, "SIFT")
kp2, des2 = findKeyPoints(img2, "SIFT")

findMatches(img1, img2, kp1, kp2, des1, des2)
