import cv2 as cv
import numpy as np
import statistics

# ----------------------------------------------------------------------------
#
# This file contains method and algorithms for feature detection and matching.
#
# ----------------------------------------------------------------------------


def findKeyPoints(frame, method="ORB", numOfKeypoints=100):
    # Check if method is a valid value
    methods = ["ORB", "SIFT", "FAST"]
    if method not in methods:
        raise ValueError("Invalid method type. Expected one of: %s" % methods)

    # Convert frame to grayscale
    frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
    # frame_gray = cannyEdgeDetecion(frame)

    # ORB - Oriented FAST and Rotated BRIEF
    if method == methods[0]:
        orb = cv.ORB_create(nfeatures=numOfKeypoints, patchSize=10)
        kp, des = orb.detectAndCompute(frame_gray, None)

        # Convert to CV_32F format
        des = np.float32(des)
        return kp, des

    # SIFT - Scale-Invariant Feature Transform
    elif method == methods[1]:
        sift = cv.xfeatures2d.SIFT_create()
        return sift.detectAndCompute(frame_gray, None)

    # FAST for keypoints and BRIEF for desciptors
    elif method == methods[2]:
        # Find keypoints with FAST
        fast = cv.FastFeatureDetector_create()
        kp = fast.detect(frame_gray, None)

        # Find descriptors with BRIEF
        brief = cv.xfeatures2d.BriefDescriptorExtractor_create()
        kp, des = brief.compute(frame_gray, kp)
        return kp, des


# Find matches between two datasets of keypoints and descriptors for two images
def findMatches(img1, img2, kp1, kp2, des1, des2):
    if len(kp1) == 0 or len(kp2) == 0 or len(des1) == 0 or len(des2) == 0:
        return []

    # FLANN matcher parameters
    FLANN_INDEX_KDTREE = (
        0  # Index for KDTREE algorithm used with floating point descriptors
    )
    FLANN_INDEX_LSH = 1  # Index for LSH algorithm used with binary desciptors
    RATIO_THRESHOLD = 0.7  # Threshold ratio value for a good match

    index_params = dict(
        algorithm=FLANN_INDEX_LSH,
        table_number=12,
        key_size=10,
        multi_probe_level=1,
    )
    search_params = dict(checks=100)

    # Create a FLANN matcher object and use knnMatch to find matches
    flann = cv.FlannBasedMatcher(index_params, search_params)

    if len(kp1) == 1 or len(kp2) == 1 or len(des1) == 1 or len(des2) == 1:
        return flann.knnMatch(des1, des2, k=1)

    matches = flann.knnMatch(
        np.asarray(des1, np.float32), np.asarray(des2, np.float32), k=2
    )

    # Save the good matches in a new array
    good_matches = []

    # Lowe's ratio test for good matches
    for m1, m2 in matches:
        # If the matches are far enough from each other, they are considered a good match
        if m1.distance < RATIO_THRESHOLD * m2.distance:
            good_matches.append([m1])

    return good_matches


# Helper method to find the distance between two keypoints in a match
def distanceBetweenMatch(match, kp1, kp2):
    # Index of keypoints to match object
    index1 = match[0].queryIdx
    index2 = match[0].trainIdx

    # Find the distance between the keypoints in the Y-direction
    return kp2[index2].pt[1] - kp1[index1].pt[1]


# Find the distance between two keypoints in a match
def distanceBetweenMatches(matches, kp1, kp2):
    if len(matches) == 0:
        return 0

    # Save distance as average in Y-direction
    totalDistance = 0

    for m in matches:
        # Calculate distances between keypoints based on index
        totalDistance += distanceBetweenMatch(m, kp1, kp2)

    # Return the average distance
    return totalDistance / len(matches)


# Filter the matches based on direction
def filterMatchesByDirection(matches, kp1, kp2):
    # Keypoints must be of same length
    # if len(kp1) != len(kp2):
    #     raise ValueError("Keypoint arrays are not of the same length")

    # Filter out the best n-matches
    # NUM_OF_MATCHES = 5
    # matches = matches[-NUM_OF_MATCHES:]

    # Minimum distance for a match to be considered an acceptable match
    DISTANCE_THRESHOLD = 0.5

    # Save matches in respective arrays based on direction
    matchesUpwards = []
    matchesDownwards = []

    # Find all distances to find average distance moved
    for m in matches:
        # Find the distance between the keypoints in the Y-direction
        distance = distanceBetweenMatch(m, kp1, kp2)
        # print("-- Distance: ", distance)

        # Index of keypoints to match object
        index1 = m[0].queryIdx
        index2 = m[0].trainIdx

        x1, y1 = kp1[index1].pt[0], kp1[index1].pt[1]
        x2, y2 = kp2[index2].pt[0], kp2[index2].pt[1]
        x1, y1 = round(x1, 1), round(y1, 1)
        x2, y2 = round(x2, 1), round(y2, 1)
        x1, y1 = str(x1), str(y1)
        x2, y2 = str(x2), str(y2)

        print("kp1: (" + x1 + ", " + y1 + ") - kp2: (" + x2 + ", " + y2 + ")")

        # Ignore matches below the distance threshold
        if abs(distance) <= DISTANCE_THRESHOLD:
            continue

        # Save match based on distance which determines the direction
        if distance > 0:
            matchesUpwards.append(m)
        else:
            matchesDownwards.append(m)

    return matchesDownwards, matchesUpwards


# Filter matches based on mean value and standard deviation of all distances
def filterMatchesByDeviation(matches, kp1, kp2):
    if len(matches) <= 3:
        return matches

    matchesFiltered = []
    distances = []

    # Find all distances between matches
    for m in matches:
        # Find the distance between the keypoints in the Y-direction
        distance = distanceBetweenMatch(m, kp1, kp2)

        # Save distances for later calculations
        distances.append(distance)

    # Calculate the mean value of the distances
    mean = statistics.mean(distances)

    # Calculate the standard deviation of the distances
    stdev = statistics.pstdev(distances)

    for i in range(0, len(distances)):
        distance = distances[i]

        # Check if distance is within 1 standard deviation of the mean value
        # A distance outside of this range is considered a bad match and should be filtered out
        if mean - stdev < distance or mean + stdev > distance:
            matchesFiltered.append(matches[i])

    return matchesFiltered


# Helper method that will filter matches with both direction and standard deviation
def filterMatches(matches, kp1, kp2, direction="Up"):
    # Possible direction for argument
    directions = ["up", "down"]

    # Remove any capitalized letters before checking
    direction = direction.lower()

    # Argument check
    if direction != directions[0] and direction != directions[1]:
        raise ValueError(
            "Direction argument must be a string containing one of these values: %s"
            % directions
        )

    # First filter matches based on direction
    matchesDownwards, matchesUpwards = filterMatchesByDirection(matches, kp1, kp2)

    # Filter matches with standard deviation calculation
    if direction == directions[0]:
        return filterMatchesByDeviation(matchesUpwards, kp1, kp2)

    return filterMatchesByDeviation(matchesDownwards, kp1, kp2)