import numpy as np
import cv2 as cv
import sys

# Fix path when importing files from different folder
sys.path.append("./")
from utils import rescaleFrame
from image import cannyEdgeDetecion, findKeyPoints, findMatches, distanceBetweenMatches

cap = cv.VideoCapture("./Resources/fouling.mp4")
cap.set(cv.CAP_PROP_POS_FRAMES, 1100)

# take first frame of the video
# _, firstFrame = cap.read()
# _, secondFrame = cap.read()

fourcc = cv.VideoWriter_fourcc(*"DIVX")

# params for ShiTomasi corner detection
feature_params = dict(
    maxCorners=50,  # How many pts. to locate
    qualityLevel=0.05,  # b/w 0 & 1, min. quality below which everyone is rejected
    minDistance=7,  # Min eucledian distance b/w corners detected
    blockSize=9,
)  # Size of an average block for computing a derivative covariation matrix over each pixel neighborhood

# Parameters for lucas kanade optical flow
lk_params = dict(
    winSize=(15, 15),  # size of the search window at each pyramid level
    maxLevel=2,  #  0, pyramids are not used (single level), if set to 1, two levels are used, and so on
    criteria=(cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT, 10, 0.03),
)

""" Criteria : Termination criteria for iterative search algorithm.
    after maxcount { Criteria_Count } : no. of max iterations.
    or after { Criteria Epsilon } : search window moves by less than this epsilon """


# Take first frame and find corners in it
ret, old_frame = cap.read()
# old_gray = cv.cvtColor(old_frame, cv.COLOR_BGR2GRAY)
old_edge = cannyEdgeDetecion(old_frame)
p0 = cv.goodFeaturesToTrack(
    old_edge, mask=None, **feature_params
)  # use goodFeaturesToTrack to find the location of the good corner.

# Create a mask image for drawing purposes filed with zeros
mask = np.zeros_like(old_frame)

y = 0
is_begin = True  # To save the output video
count = 1  # for the frame count
n = 10  # Frames refresh rate for feature generation

distanceTraveled_old = 0
distanceTraveled_new = 0
speed_old = 0
speed_new = 0
refreshIndex = 0

while True:
    ret, frame = cap.read()
    if frame is None:
        break
    processed = frame

    # if count == 2:
    #     break

    # Saving the Video
    if is_begin:
        h, w, _ = processed.shape
        # out = cv.VideoWriter(output_file, fourcc, 30, (w, h), True)
        is_begin = False

    # Convert to Grey Frame
    # frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
    frame_edge = cannyEdgeDetecion(frame)

    if count % n == 0:  # Refresh the tracking features after every n frames
        # cv.imwrite("img/r{0:05d}.jpg".format(y), img)
        y += 1
        ret, old_frame = cap.read()
        # old_gray = cv.cvtColor(old_frame, cv.COLOR_BGR2GRAY)
        old_edge = cannyEdgeDetecion(old_frame)
        p0 = cv.goodFeaturesToTrack(old_edge, mask=None, **feature_params)
        mask = np.zeros_like(old_frame)

        refreshIndex = count

    # calculate optical flow
    p1, st, err = cv.calcOpticalFlowPyrLK(old_edge, frame_edge, p0, None, **lk_params)

    # Select good points
    good_new = p1[st == 1]
    good_old = p0[st == 1]

    # print(good_new)

    # draw the tracks
    for i, (new, old) in enumerate(zip(good_new, good_old)):
        a, b = new.ravel()  # tmp new value
        c, d = old.ravel()  # tmp old value

        distanceTraveled_new += abs(b - d)

        # draws a line connecting the old point with the new point
        mask = cv.line(mask, (a, b), (c, d), (0, 0, 255), 3)

        # draws the new point
        frame = cv.circle(frame, (a, b), 3, (0, 255, 0), -1)

    distanceTraveled_new = distanceTraveled_new / len(good_new)
    speed = distanceTraveled_new
    MIN_DIFFERENCE = 3
    SPEED_THRESHOLD = 1.5

    if abs(distanceTraveled_old - distanceTraveled_new) >= MIN_DIFFERENCE:
        speed_new = (distanceTraveled_new + distanceTraveled_old) / 2

    else:
        speed_new = distanceTraveled_new

    speed_new = (speed_new + speed_old) / 2

    # if distanceTraveled_new <= 5:
    #     # print(distanceTraveled_new)

    if refreshIndex == count or refreshIndex == count + 1:
        speed_new = speed_old

    if speed_new < SPEED_THRESHOLD:
        speed_new = 0

    # For next frame!
    distanceTraveled_old = distanceTraveled_new
    distanceTraveled_new = 0
    speed_old = speed_new

    gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
    sift = cv.SIFT_create()
    kp = sift.detect(gray, None)
    keypoints = cv.drawKeypoints(
        frame, kp, frame, flags=cv.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS
    )

    frame_edge = cv.bitwise_not(frame_edge)
    # Change edge colors
    edge_rgb = cv.cvtColor(frame_edge, cv.COLOR_GRAY2RGB)
    edge_rgb *= np.array((0, 0, 1), np.uint8)
    # print(edge_rgb.shape)

    fgMask = cv.createBackgroundSubtractorKNN().apply(frame)

    cv.rectangle(frame, (10, 2), (100, 20), (255, 255, 255), -1)
    cv.putText(
        frame,
        str(cap.get(cv.CAP_PROP_POS_FRAMES)),
        (15, 15),
        cv.FONT_HERSHEY_SIMPLEX,
        0.5,
        (0, 0, 0),
    )

    frame_masked = np.bitwise_or(frame, frame_edge[:, :, np.newaxis])
    img = cv.add(frame, mask)

    cv.rectangle(frame, (40, 4), (670, 110), (0, 0, 0), -1)
    cv.putText(
        frame,
        "Speed: " + "{0:.2g}".format(speed_new),
        (60, 80),
        cv.FONT_HERSHEY_TRIPLEX,
        3,
        (0, 0, 255),
        2,
    )

    # cv.imshow("FG Mask", rescaleFrame(fgMask, 0.4))

    cv.imshow("frame", rescaleFrame(frame, 0.4))
    cv.imshow("masked", rescaleFrame(img, 0.4))
    cv.imshow("keypoints", rescaleFrame(keypoints, 0.4))
    k = cv.waitKey(30) & 0xFF

    # Show the Output
    if k == 27:
        # cv.imshow("", img)
        break

    # Now update the previous frame and previous points
    old_frame = frame.copy()
    p0 = good_new.reshape(-1, 1, 2)

    count += 1

# release and destroy all windows
cv.destroyAllWindows()
cap.release()